import vanilla
from botocore.exceptions import NoCredentialsError, EndpointConnectionError
from django.db.models import Count, Min, OuterRef, Subquery
from django.db.models import ExpressionWrapper, F, DurationField
from django.shortcuts import render
from django.views.generic import ListView, TemplateView
from otree.models import Session, Participant
from otree.views.export import get_export_response
# BLOCK FOR MTURK HITS
from otree.views.mturk import get_mturk_client
from otree_tools.models import EnterEvent, ExitEvent, FocusEvent
import json
from django.http import HttpResponse, StreamingHttpResponse
from .export import export_wide
from django.template import loader
import csv
from datetime import timedelta
from django.core.urlresolvers import reverse
from . import __version__ as otree_tools_version


# import otree_tools.forms as forms
# END OF BLOCK
# BLOCK FOR TESTING JSON THINGS
# from .forms import (UpdateExpirationForm)
# END OF BLOCK


# TIME STAMPS VIEWS FOR TRACKING_TIME and TRACKING_FOCUS
class EnterExitMixin:
    def get_queryset(self):
        # when we request num_exits>0 we get only those enter events which have some corresponding
        # exit events. we should sort them based on their exit type because form is submitted before or at the
        # same time as page is unloaded etc.
        earliest_exit = ExitEvent.objects.filter(enter_event=OuterRef('pk')).order_by('exit_type', 'timestamp')
        subquery_earliest = Subquery(earliest_exit.values('pk')[:1])

        tot_enter_events = EnterEvent.objects. \
            annotate(num_exits=Count('exits')). \
            filter(num_exits__gt=0). \
            annotate(
            early_exits=Min('exits__timestamp'),
            early_exit_pk=subquery_earliest,
            timediff=ExpressionWrapper(F('early_exits') - F('timestamp'),
                                       output_field=DurationField())
        )
        # TODO: later on think about efficiency of looping through all foo_display and replace exittype to

        for i in tot_enter_events:
            # for some weird reasons some random time diffs are not calculated at all (like None).
            if i.timediff is None:
                i.timediff = i.early_exits - i.timestamp

            i.exittype = ExitEvent.objects.get(pk=i.early_exit_pk).get_exit_type_display()

        return tot_enter_events


class PaginatedListView(ListView):
    navbar_active_tag = None
    export_link_name = None

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        c['nbar'] = self.navbar_active_tag
        curpage_num = c['page_obj'].number
        paginator = c['paginator']
        epsilon = 3
        c['allowed_range'] = range(max(1, curpage_num - epsilon), min(curpage_num + epsilon, paginator.num_pages) + 1)
        if self.export_link_name:
            c['export_link'] = reverse(self.export_link_name)
        return c


class EnterExitEventList(EnterExitMixin, PaginatedListView):
    template_name = 'otree_tools/all_timespent_list.html'
    url_name = 'time_spent_timestamps'
    url_pattern = r'^time_spent_per_page/$'
    display_name = 'Time spent per page [otree-tools]'
    context_object_name = 'timestamps'
    paginate_by = 50
    navbar_active_tag = 'time'
    export_link_name = 'enterexit_csv_export'


class FocusEventList(PaginatedListView):
    template_name = 'otree_tools/focus_event_list.html'
    url_name = 'focus_events'
    url_pattern = r'^focus_events/$'
    display_name = 'Focus/unfocus events [otree-tools]'
    context_object_name = 'focusevents'
    queryset = FocusEvent.objects.all()
    paginate_by = 50
    navbar_active_tag = 'focus'
    export_link_name = 'streaming_focus_csv'


class Echo:
    def write(self, value):
        return value


class StreamingCSVExport(ListView):
    response_class = StreamingHttpResponse
    content_type = 'text/csv'

    def iter_items(self, items, pseudo_buffer):
        writer = csv.DictWriter(pseudo_buffer, fieldnames=self.get_headers())
        yield writer.writerow(dict((fn, fn) for fn in self.get_headers()))

        for item in items:
            yield writer.writerow(self.get_data(item))

    def get(self, request, *args, **kwargs):
        response = StreamingHttpResponse(
            streaming_content=(self.iter_items(self.get_queryset(), Echo())),
            content_type='text/csv',
        )
        response['Content-Disposition'] = 'attachment;filename={}'.format(self.filename)
        return response


class StreamingFocusCSV(StreamingCSVExport):
    # todo: deal with coinciding time stamps of PageUnload and Form Submitted events
    url_name = 'streaming_focus_csv'
    url_pattern = r'^streaming_focusevent_csv_export/$'
    filename = 'focus_data.csv'

    def get_queryset(self):
        return FocusEvent.objects.all()

    def get_headers(self):
        return ['session_code', 'participant_code', 'app_name', 'round_number', 'page_name', 'timestamp',
                'event_desc_type', ]

    def get_data(self, item):
        #  unfortunately we can't make it more compact because reverse relations cannot be obtained with
        # general relations (or at least im too lazy to figure out how
        return {'session_code': item.participant.session.code,
                'participant_code': item.participant.code,
                'app_name': item.app_name,
                'round_number': item.player.round_number,
                'page_name': item.page_name,
                'timestamp': item.timestamp,
                'event_desc_type': item.event_desc_type}


class StreamingEnterCSV(EnterExitMixin, StreamingCSVExport):
    url_name = 'enterexit_csv_export'
    url_pattern = r'^enterexit_csv_export/$'
    filename = 'enter_exit_data.csv'

    def get_headers(self):
        return ['session_code', 'participant_code', 'app_name', 'round_number', 'page_name',
                'entrance_timestamp', 'exit_timestamp', 'duration', 'exit_type'
                ]

    def get_data(self, item):
        return {'session_code': item.participant.session.code,
                'participant_code': item.participant.code,
                'app_name': item.app_name,
                'round_number': item.player.round_number,
                'page_name': item.page_name,
                'entrance_timestamp': item.timestamp,
                'exit_timestamp': item.early_exits,
                'duration': item.timediff,
                'exit_type': item.exittype}


# END OF TIME STAMPS BLOCK


# Welcoming page
class HomeView(TemplateView):
    template_name = 'otree_tools/home.html'
    url_name = 'otree_tools_home'
    url_pattern = r'^otree_tools/$'
    display_name = 'otree-tools Home page'
    navbar_active_tag = 'home'

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        c['nbar'] = self.navbar_active_tag
        c['version'] = otree_tools_version
        return c


# End of welcoming page

# Dealin with individual session export - a list of all sessions goes first
class AllSessionsList(PaginatedListView):
    template_name = 'otree_tools/all_session_list.html'
    url_name = 'individual_sessions_export'
    url_pattern = r'^individual_sessions_export/$'
    display_name = 'Exporting data from individual sessions'
    navbar_active_tag = 'export'
    queryset = Session.objects.all()
    context_object_name = 'sessions'
    paginate_by = 10


# END OF BLOCK:: Dealin with individual session export - a list of all sessions goes first
# Block dealing with exporting participant vars #


class PVarsSessionsList(ListView):
    template_name = 'otree_tools/pvar_session_list.html'
    url_name = 'sessions_pvar_export'
    url_pattern = r'^sessions_pvar_export/$'
    display_name = 'Exporting Participant Vars [otree-tools]'
    model = Session
    context_object_name = 'sessions'


class UniversalEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (set, range)):
            return list(obj)
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            try:
                return str(obj)
            # the last resort - if built-in __str__ of the object causes an error we fail silently
            except:
                return ''


class PVarsMixin(object):
    def get_session(self):
        return Session.objects.get(pk=self.kwargs.get('pk'))

    def get_queryset(self):
        headers_to_add = ['session.code', 'player.code']
        session = self.get_session()
        data = Participant.objects.filter(session=session).values('vars', 'code', 'session__code')
        # todo: sort the headers
        keys = sorted(list(set().union(*(d['vars'].keys() for d in data))), key=lambda s: s.lower())
        self.heads = headers_to_add + keys
        q = list()
        for i in data:
            dict_to_row = json.loads(json.dumps(i['vars'], cls=UniversalEncoder))
            row = [i['session__code'], i['code']] + [dict_to_row.get(key, '') for key in keys]
            q.append(row)

        return q


class ListPVarsView(PVarsMixin, PaginatedListView):
    template_name = 'otree_tools/pvars_list.html'
    url_name = 'pvars_list'
    url_pattern = r'^session/(?P<pk>[a-zA-Z0-9_-]+)/pvars/$'
    model = Participant
    context_object_name = 'pvars'
    paginate_by = 20
    navbar_active_tag = 'export'
    export_link_name = None

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        session = self.get_session()
        c['session'] = session
        c['heads'] = self.heads
        c['export_link'] = reverse('pvars_csv_export', kwargs={'pk': session.pk})
        return c


class PVarsCSVExport(PVarsMixin, TemplateView):
    url_name = 'pvars_csv_export'
    url_pattern = r'^session/(?P<pk>[a-zA-Z0-9_-]+)/pvars_csv_export/$'
    response_class = HttpResponse
    content_type = 'text/csv'

    def get(self, request, *args, **kwargs):
        response = HttpResponse(content_type='text/csv')
        session_code = self.get_session().code
        filename = '{}_participant_vars.csv'.format(session_code)
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(filename)

        pvars = self.get_queryset()
        writer = csv.writer(response)
        writer.writerow(self.heads)

        for p in pvars:
            writer.writerow(p)

        return response


# END OF Block dealing with exporting participant vars #


def check_if_deletable(h):
    if (h['HITStatus'] == 'Reviewable' and
                    h['NumberOfAssignmentsCompleted'] +
                    h['NumberOfAssignmentsAvailable'] == h['MaxAssignments']):
        h['Deletable'] = True
    return h


NO_CRED_ERR_CODE = 0
NO_CONN_ERR_CODE = 1


class MturkClient(object):
    client = None
    errors = None

    def get_errors(self):
        if self.errors == NO_CONN_ERR_CODE:
            return 'Sorry, there is no connection to Amazon mTurk web-site. Check your internet connection.'
        if self.errors == NO_CRED_ERR_CODE:
            return 'Sorry, I cannot find your Amazon credentials. Check your environment variables.'

    def __init__(self, use_sandbox=True):
        try:
            self.client = get_mturk_client(use_sandbox=use_sandbox)
            self.client.get_account_balance()
        except NoCredentialsError:
            self.client = None
            self.errors = NO_CRED_ERR_CODE
        except EndpointConnectionError:
            self.client = None
            self.errors = NO_CONN_ERR_CODE


class SpecificSessionDataView(vanilla.TemplateView):
    def get(self, request, *args, **kwargs):
        session_code = kwargs['session_code']
        response, file_extension = get_export_response(
            request, session_code)
        export_wide(response, file_extension, session_code=session_code)
        return response


class HitsList(vanilla.TemplateView):
    template_name = 'otree_tools/hits_list.html'
    url_name = 'hits_list'
    url_pattern = r'^hits_list/$'
    display_name = 'mTurk HITs'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        mturk = MturkClient()
        client = mturk.client

        if client is not None:
            balance = client.get_account_balance()['AvailableBalance']
            hits = client.list_hits()['HITs']
            for h in hits:
                h = check_if_deletable(h)
            context['balance'] = balance
            context['hits'] = hits
        else:
            context['mturk_errors'] = mturk.get_errors()

        return context


class AssignmentListView(vanilla.TemplateView):
    template_name = 'otree_tools/assignments_list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        current_hit_id = self.kwargs.get('HITId')
        mturk = MturkClient()
        client = mturk.client
        if client is not None:
            cur_hit = check_if_deletable(client.get_hit(HITId=current_hit_id).get('HIT'))
            context['hit'] = cur_hit
            assignments = client.list_assignments_for_hit(HITId=current_hit_id)['Assignments']
            submitted_assignments = bool(
                {'Submitted', 'Rejected'} & set([a['AssignmentStatus'] for a in assignments]))
            context['assignments'] = assignments
            context['submitted_assignments'] = submitted_assignments
        else:
            context['mturk_errors'] = mturk.get_errors()
        return context

#
#
# class SendSomethingView(vanilla.FormView):
#     HITId = None
#     AssignmentId = None
#     WorkerId = None
#     Assignment = None
#
#     def dispatch(self, request, *args, **kwargs):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             self.assignment = client.get_assignment(AssignmentId=kwargs['AssignmentID'])['Assignment']
#             self.AssignmentId = self.assignment['AssignmentId']
#             self.WorkerId = self.assignment['WorkerId']
#             self.HITId = self.assignment['HITId']
#         else:
#             self.context['mturk_errors'] = mturk.get_errors()
#         return super().dispatch(request, *args, **kwargs)
#
#     def get_context_data(self, **kwargs):
#         context = super().get_context_data(**kwargs)
#         context['assignment'] = self.assignment
#         return context
#
#     def get_success_url(self):
#         return reverse('assignments_list', kwargs={'HITId': self.HITId})
#
#
# class SendMessageView(SendSomethingView):
#     form_class = forms.SendMessageForm
#     template_name = 'otree_tools/send_message.html'
#
#     def form_valid(self, form):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             sending_message = client.notify_workers(
#                 Subject=form.cleaned_data['subject'],
#                 MessageText=form.cleaned_data['message_text'],
#                 WorkerIds=[self.WorkerId, ]
#             )
#         return super().form_valid(form)
#
#
# class SendBonusView(SendSomethingView):
#     template_name = 'otree_tools/send_bonus.html'
#     form_class = forms.SendBonusForm
#
#     def get_form(self, data=None, files=None, **kwargs):
#         max_bon = 100
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.list_bonus_payments(
#                 HITId=self.HITId,
#                 MaxResults=100,
#             )
#             bs = response['BonusPayments']
#             today = datetime.utcnow().date()
#             start = datetime(today.year, today.month, today.day, tzinfo=tz.tzutc())
#             recent_bs = [float(i['BonusAmount']) for i in bs if i['GrantTime'] > start]
#             tot_bon = sum([i for i in recent_bs])
#             max_bon = max(0, 100 - tot_bon)
#
#         cls = self.get_form_class()
#         return cls(data=data, files=files, max_bonus=max_bon)
#
#     def form_valid(self, form):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.send_bonus(
#                 WorkerId=self.WorkerId,
#                 BonusAmount=str(form.cleaned_data['bonus_amount']),
#                 AssignmentId=self.AssignmentId,
#                 Reason=form.cleaned_data['reason'],
#             )
#         return super().form_valid(form)
#
#
# class DeleteHitView(vanilla.View):
#     def get(self, request, *args, **kwargs):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             cur_hit = check_if_deletable(client.get_hit(HITId=self.kwargs['HITId']).get('HIT'))
#             if cur_hit.get('Deletable'):
#                 response = client.delete_hit(HITId=cur_hit['HITId'])
#
#         return HttpResponseRedirect(reverse_lazy('hits_list'))
#
#
# class UpdateExpirationView(vanilla.FormView):
#     back_to_HIT = None
#     form_class = UpdateExpirationForm
#     template_name = 'otree_tools/update_expiration.html'
#     HITId = None
#     HIT = None
#
#     def get_form(self, data=None, files=None, **kwargs):
#         cls = self.get_form_class()
#         return cls(data=data, files=files, initial={'expire_time': self.HIT['Expiration']})
#
#     def get_success_url(self):
#         if self.back_to_HIT:
#             return reverse('assignments_list', kwargs={'HITId': self.HITId})
#         else:
#             return reverse_lazy('hits_list')
#
#     def dispatch(self, request, *args, **kwargs):
#
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             self.HITId = kwargs['HITId']
#             self.HIT = client.get_hit(HITId=self.HITId).get('HIT')
#         else:
#             self.context['mturk_errors'] = mturk.get_errors()
#         return super().dispatch(request, *args, **kwargs)
#
#     def form_valid(self, form):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.update_expiration_for_hit(
#                 HITId=self.HITId,
#                 ExpireAt=0  # form.cleaned_data['expire_time']
#             )
#             response = client.update_expiration_for_hit(
#                 HITId=self.HITId,
#                 ExpireAt=form.cleaned_data['expire_time']
#             )
#         return super().form_valid(form)
#
#
# class ExpireHitView(vanilla.View):
#     back_to_HIT = None
#
#     def get(self, request, *args, **kwargs):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.update_expiration_for_hit(
#                 HITId=self.kwargs['HITId'],
#                 ExpireAt=0,
#             )
#         if self.back_to_HIT:
#             return HttpResponseRedirect(reverse('assignments_list', kwargs={'HITId': self.kwargs['HITId']}))
#         else:
#             return HttpResponseRedirect(reverse_lazy('hits_list'))
#
#
# class ApproveAssignmentView(vanilla.FormView):
#     form_class = forms.ApproveAssignmentForm
#     template_name = 'otree_tools/approve_assignment.html'
#     success_url = reverse_lazy('hits_list')
#
#     def get(self, request, *args, **kwargs):
#         return super().get(request, *args, **kwargs)
#
#     def form_valid(self, form):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.approve_assignment(
#                 AssignmentId=self.kwargs['AssignmentID'],
#                 RequesterFeedback=form.cleaned_data['message_text'],
#                 OverrideRejection=True,
#             )
#         return super().form_valid(form)
#
#
# class RejectAssignmentView(vanilla.FormView):
#     form_class = forms.RejectAssignmentForm
#     template_name = 'otree_tools/reject_assignment.html'
#     success_url = reverse_lazy('hits_list')
#
#     def form_valid(self, form):
#         mturk = MturkClient()
#         client = mturk.client
#         if client is not None:
#             response = client.reject_assignment(
#                 AssignmentId=self.kwargs['AssignmentID'],
#                 RequesterFeedback=form.cleaned_data['message_text'],
#             )
#         return super().form_valid(form)



import floppyforms as forms




class SendBonusForm(forms.Form):
    def __init__(self, max_bonus, *args,**kwargs):
        super().__init__(*args,**kwargs)
        self.fields['bonus_amount']=forms.FloatField(min_value=0, max_value=max_bonus)
        self.fields['reason'] = forms.CharField(widget=forms.Textarea)


class SendMessageForm(forms.Form):
    subject = forms.CharField(max_length=100)
    message_text = forms.CharField(widget=forms.Textarea)


class ApproveAssignmentForm(forms.Form):
    message_text = forms.CharField(widget=forms.Textarea, required=False)


class RejectAssignmentForm(forms.Form):
    message_text = forms.CharField(widget=forms.Textarea, required=False)


from datetimewidget.widgets import DateTimeWidget, DateWidget, TimeWidget

class UpdateExpirationForm(forms.Form):
    expire_time = forms.DateTimeField(widget=DateTimeWidget(usel10n=True, bootstrap_version=3))



# -*- coding: utf-8 -*-

from django.forms import MultiValueField, ChoiceField

from .widgets import RadioGridWidget


class RadioGridFormField(MultiValueField):

    def __init__(self, rows, values, *args, **kwargs):
        kwargs['widget'] = RadioGridWidget(rows, values)
        kwargs['fields'] = [ChoiceField(choices=values, required=False) for _ in rows]
        super(RadioGridFormField, self).__init__(*args, **kwargs)

    def compress(self, data_list):
        return '' if data_list is None else ','.join(data_list)



from django.db import models
from otree.models import Participant
from django.db import transaction
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey

EXITTYPES = [(0, 'form submitted'), (1, 'page unloaded'), (2, 'client disconnected')]
FOCUS_EVENT_TYPES = (
(0, 'Page shown'), (1, 'Visibility: off'), (2, 'Focus: off'), (3, 'Visibility: on'), (4, 'Focus: on'), (5,'Form submitted'))


# There are 3 different scenarios how a client may exit the page.
# 1. He can submit the form by clicking next (or in oTree any other button because the entire page
# is wrapped into form tags
# 2. He can refresh the page (clicking F5 or command+R)
# 3. The browser can be closed (by accident or intentionally)

# The Db structure is the following:
# Enter timestamp is linked to participant, and it also contains info on page. All timestamps correspond on time
# when a participant connects to a channel on page load.

# since there are several exit timestamp events that can be linked to one single enter event, that means that they
# should be connected to an enter timestamp via ForeignKey.
# How the connection takes place:
# The Enter event has  a boolean attribute Closed, initially set to false.
# When the client disconnects it looks for all not closed enter events for this participant-page pair, and closes them
# As a safeguard, the same closing happens everytime a client connects to the page, and before a new open Enter event is
# created.

# The aggregation for stats purposes is done the following way:
# * we choose all enter events for the specific participant and page
# * collect the earliest exit event for each enter event (via subquery and aggregation)
# * annotate the query via F and ExpressionWrapper to calculate the difference between this earlist exit and its parent
# Enter event
# ...
# PROFIT!!
class OpenEnterEventManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(closed=False)

    def close_all(self, participant, page_name):
        q = self.get_queryset().filter(participant=participant, page_name=page_name)
        q.update(closed=True)


class GeneralEvent(models.Model):
    class Meta:
        get_latest_by = 'timestamp'
        abstract = True

    page_name = models.CharField(max_length=1000)
    participant = models.ForeignKey(to=Participant, related_name="%(app_label)s_%(class)s_events")
    app_name = models.CharField(max_length=1000)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    player_id = models.PositiveIntegerField()
    player = GenericForeignKey('content_type', 'player_id')
    timestamp = models.DateTimeField()


class EnterEvent(GeneralEvent):
    class Meta(GeneralEvent.Meta):
        base_manager_name = 'objects'

    objects = models.Manager()
    opened = OpenEnterEventManager()
    # do we need closed? the disconnect method does it too late
    closed = models.BooleanField(default=False)

    def __str__(self):
        return 'id: {}, Enter: {}; time: {}; closed: {}'.format(self.pk, self.page_name, self.timestamp, self.closed)


class ExitEvent(models.Model):
    class Meta:
        get_latest_by = 'timestamp'

    enter_event = models.ForeignKey(to=EnterEvent, related_name='exits')
    timestamp = models.DateTimeField()
    exit_type = models.IntegerField(choices=EXITTYPES)

    def __str__(self):
        return 'id: {}, time: {}; for ENRANCE event {}. Type: {} '.format(self.pk,
                                                                          self.timestamp,
                                                                          self.enter_event.timestamp,
                                                                          self.get_exit_type_display())


class FocusEvent(GeneralEvent):
    event_desc_type = models.CharField(max_length=1000)
    event_num_type = models.IntegerField(choices=FOCUS_EVENT_TYPES)



from django.conf.urls import url, include

from otree_tools import views as v
from serializer_ext.views import DownloadJson

urlpatterns = [url(v.ListPVarsView.url_pattern, v.ListPVarsView.as_view(),
                   name=v.ListPVarsView.url_name),
               url(v.PVarsCSVExport.url_pattern, v.PVarsCSVExport.as_view(),
                   name=v.PVarsCSVExport.url_name),
               url(v.StreamingEnterCSV.url_pattern, v.StreamingEnterCSV.as_view(),
                   name=v.StreamingEnterCSV.url_name),
               url(v.AllSessionsList.url_pattern, v.AllSessionsList.as_view(),
                   name=v.AllSessionsList.url_name),
               url(v.FocusEventList.url_pattern, v.FocusEventList.as_view(),
                   name=v.FocusEventList.url_name),
               url(v.EnterExitEventList.url_pattern, v.EnterExitEventList.as_view(),
                   name=v.EnterExitEventList.url_name),
               url(r'^session_data/(?P<session_code>.*)/(?P<filetype>.*)/$', v.SpecificSessionDataView.as_view(),
                   name='session_data'),
               url(r'^download_json/(?P<session_code>\w+)$', DownloadJson.as_view(),
                   name='download_json'),
               url(v.StreamingFocusCSV.url_pattern, v.StreamingFocusCSV.as_view(),
                   name=v.StreamingFocusCSV.url_name),
               ]



from otree.models_concrete import PageCompletion
from otree_tools.models import EnterEvent, FocusEvent
from django.db.models import Count, Min, Sum, Q, ExpressionWrapper, F, DurationField
from datetime import timedelta
import errno
import shutil
import logging


def copy(src, dest):
    try:
        shutil.copytree(src, dest)
        return True
    except OSError as e:
        # If the error was caused because the source wasn't a directory
        if e.errno == errno.ENOTDIR:
            shutil.copy(src, dest)
            return True
        else:
            logger = logging.getLogger('otree-tools')
            logger.error('Directory not copied. Error: {}'.format(e))
            return False


def get_seconds_per_page(player, page_name):
    participant = player.participant
    subsession_pk = player.subsession.pk
    try:
        page_info = PageCompletion.objects.get(participant=participant, subsession_pk=subsession_pk,
                                               page_name=page_name)
    except PageCompletion.DoesNotExist:
        raise ValueError('Cannot find info about this page')
    return page_info.seconds_on_page


def get_time_per_page(player, page_name):
    tot_enter_events = EnterEvent.objects.filter(player_id=player.pk, participant=player.participant,
                                                 page_name=page_name
                                                 ).annotate(num_exits=Count('exits')).filter(num_exits__gt=0)
    if tot_enter_events.exists():
        b = tot_enter_events.annotate(
            early_exits=Min('exits__timestamp'),
            timediff=ExpressionWrapper(F('early_exits') - F('timestamp'),
                                       output_field=DurationField())
        ).aggregate(sum_diff=Sum('timediff'))

        sum_diff = b['sum_diff']

        return sum_diff.total_seconds()


def _aggregate_focus_time(player, page_name, focus_on=True):
    relfocuses = FocusEvent.objects.filter(player_id=player.pk, participant=player.participant,
                                           page_name=page_name).order_by('-timestamp')
    if not relfocuses.exists():
        return

    exit_types = [1, 2, 5]
    enter_types = [0, 3, 4]
    if relfocuses.first().event_num_type not in exit_types:
        return
    if relfocuses.last().event_num_type not in enter_types:
        return
    for i in relfocuses:
        if i.event_num_type in exit_types:
            i.general_type = 0
        else:
            i.general_type = 1

    tot_focus_on = tot_focus_off = timedelta()

    for index, obj in enumerate(relfocuses):
        if index > 0:
            previous = relfocuses[index - 1]
            if obj.general_type != previous.general_type:
                if obj.general_type == 1:
                    tot_focus_on += (previous.timestamp - obj.timestamp)
                else:
                    tot_focus_off += (previous.timestamp - obj.timestamp)
    if focus_on:
        return tot_focus_on.total_seconds()
    else:
        return tot_focus_off.total_seconds()


def get_focused_time(player, page_name):
    return _aggregate_focus_time(player, page_name)


def get_unfocused_time(player, page_name):
    return _aggregate_focus_time(player, page_name, focus_on=False)



from otree_tools.widgets import OtherSelectorWidget, divider
from django.forms.fields import MultiValueField, CharField, MultipleChoiceField, ChoiceField
from otree.common_internal import expand_choice_tuples
from django.core.exceptions import ValidationError
from django.forms import CheckboxSelectMultiple

duplicate_err_msg = 'One of the choices duplicates with the internal name for "Other" choice field'
wrong_value_msg = 'Please provide a valid name for other_value. It should look like Python variable (no spaces etc.)'
wrong_type_err_msg = 'Please provide a choices option for ListField which should be either a list or tuple of values'


class OtherFormField(MultiValueField):
    other_value = 'other_'
    other_label = 'Other'
    required = True

    def __init__(self, other_value=None, other_label=None, label='', **kwargs):

        self.choices = kwargs.pop('choices')
        if other_label:
            self.other_label = other_label
        if other_value:
            self.other_value = other_value
        # check that other_value provided is valid
        assert self.other_value.isidentifier(), wrong_value_msg
        if isinstance(self.choices, (list, tuple)) and len(self.choices) > 0:
            self.choices = list(expand_choice_tuples(self.choices))
            flat_choices = [i for i, j in self.choices]
            # check if value of choices start with other_val and divider, like 'other: SOMETHING'
            for i in flat_choices:
                assert not i.startswith(self.other_value), duplicate_err_msg
            self.choices += [(self.other_value, self.other_label), ]
            self.widget = OtherSelectorWidget(choices=self.choices, other_val=self.other_value)
        fields = (CharField(required=self.required), CharField(required=False),)
        super().__init__(fields=fields, require_all_fields=False, label=label, **kwargs)

    def compress(self, data_list):
        # this fallback is needed only if the field is blank
        if len(data_list) > 0:
            if data_list[0] == self.other_value:
                if data_list[1] in (None, ''):
                    raise ValidationError(u'Please provide the answer for "{}" field'.format(self.other_label))
                return '{}{}{}'.format(self.other_value, divider, data_list[1])
            else:
                return data_list[0]


class MultiSelectFormField(MultipleChoiceField):
    widget = CheckboxSelectMultiple

    def __init__(self, *args, **kwargs):
        self.max_choices = kwargs.pop('max_choices', None)
        self.min_choices = kwargs.pop('min_choices', None)
        super(MultiSelectFormField, self).__init__(*args, **kwargs)

    def clean(self, value):
        if not value and self.required:
            raise ValidationError(self.error_messages['required'])
        if value and self.max_choices and self.min_choices:
            if self.max_choices == self.min_choices:
                if len(value) != self.max_choices:
                    ending = '' if self.max_choices == 1 else 's'
                    raise ValidationError('You must select exactly {} choice{}.'.format(self.max_choices, ending))
        if value and self.max_choices:
            if len(value) > self.max_choices:
                ending = '' if self.max_choices == 1 else 's'
                raise ValidationError('You must select a maximum of {} choice{}.'.format(self.max_choices, ending))
        if value and self.min_choices:
            if len(value) < self.min_choices:
                ending = '' if self.min_choices == 1 else 's'
                raise ValidationError('You must select a minimum of {} choice{}.'.format(self.min_choices, ending))
        return value



import json
from datetime import datetime

from channels.generic.websockets import JsonWebsocketConsumer
from django.contrib.contenttypes.models import ContentType
from otree.models import Participant
from otree_tools.models import EnterEvent, FocusEvent
from otree.models_concrete import ParticipantToPlayerLookup


class GeneralTracker(JsonWebsocketConsumer):
    tracker_url_kwargs = r'(?P<participant_code>[a-zA-Z0-9_-]+)/(?P<page_name>[a-zA-Z0-9_-]+)/(?P<page_index>\d+)$'

    def clean_kwargs(self):
        self.participant_code = self.kwargs['participant_code']
        self.page_name = self.kwargs['page_name']
        self.page_index = self.kwargs['page_index']

    def get_participant(self):
        self.clean_kwargs()
        try:
            return Participant.objects.get(code__exact=self.participant_code)
        except Participant.DoesNotExist:
            return

    def get_player_and_app(self):
        empty_ret = (None, None, None)
        participant = self.get_participant()
        if participant:
            try:
                participant_lookup_item = participant.participanttoplayerlookup_set.get(
                    page_index=self.page_index)
                app_name = participant_lookup_item.app_name
                player_pk = participant_lookup_item.player_pk

                player_content_type = ContentType.objects.get(app_label=app_name, model='player')
                player = player_content_type.get_object_for_this_type(pk=player_pk)
                return participant, app_name, player

            except ParticipantToPlayerLookup.DoesNotExist:
                ...
        return empty_ret


class TimeTracker(GeneralTracker):
    url_pattern = (r'^/timetracker/' + GeneralTracker.tracker_url_kwargs)

    def get_unclosed_enter_event(self):
        p = self.get_participant()
        o = EnterEvent.opened.filter(participant=p)
        if o.exists():
            return o.latest()

    def receive(self, content, **kwargs):
        raw_content = json.loads(content)
        raw_time = raw_content['timestamp']
        timestamp = datetime.fromtimestamp(raw_time / 1000)
        event_type = raw_content['eventtype']
        if event_type == 'enter':
            participant, app_name, player = self.get_player_and_app()
            if participant is not None:
                EnterEvent.opened.close_all(participant, self.page_name, )

                participant.otree_tools_enterevent_events.create(page_name=self.page_name,
                                                                 timestamp=timestamp,
                                                                 player=player,
                                                                 app_name=app_name)

        if event_type == 'exit':
            exit_type = int(raw_content['exittype'])
            latest_entry = self.get_unclosed_enter_event()
            if latest_entry is not None:
                latest_entry.exits.create(timestamp=timestamp,
                                          exit_type=exit_type)

    def connect(self, message, **kwargs):
        print('Client connected to time tracker...')

    def disconnect(self, message, **kwargs):
        participant = self.get_participant()
        latest_entry = self.get_unclosed_enter_event()
        if latest_entry is not None:
            latest_entry.exits.create(timestamp=datetime.now(),
                                      exit_type=2)

            EnterEvent.opened.close_all(participant, self.page_name)


class FocusTracker(GeneralTracker):
    url_pattern = (r'^/focustracker/' + GeneralTracker.tracker_url_kwargs)

    def receive(self, content, **kwargs):
        raw_content = json.loads(content)
        raw_time = raw_content['timestamp']
        timestamp = datetime.fromtimestamp(raw_time / 1000)
        event_num_type = raw_content['event_num_type']
        event_desc_type = raw_content['event_desc_type']

        participant, app_name, player = self.get_player_and_app()
        if participant is not None:
             participant.otree_tools_focusevent_events.create(page_name=self.page_name,
                                                             timestamp=timestamp,
                                                             player=player,
                                                             app_name=app_name,
                                                             event_desc_type=event_desc_type,
                                                             event_num_type=event_num_type, )


    def connect(self, message, **kwargs):
        print('Client connected to focus tracker...')



from django.db import models
import json
from django.core import checks
from otree_tools.forms.fields import OtherFormField, MultiSelectFormField
from otree.widgets import CheckboxSelectMultiple
from otree.common_internal import expand_choice_tuples
from otree_tools.radiogrid import RadioGridField
from otree_tools.widgets import NoCheckboxCheckbox


class InnerChoiceMixin:
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name == 'choices':
            if isinstance(value, (list, tuple)) and len(value) > 0:
                self.inner_choices = list(expand_choice_tuples(value))


class OtherModelField(InnerChoiceMixin, models.CharField):
    other_value = None
    other_label = None

    def __init__(self, max_length=10000,
                 blank=False,
                 label=None,
                 *args, **kwargs):
        kwargs.update(dict(label=label, ))
        kwargs.setdefault('help_text', '')
        kwargs.setdefault('null', True)
        kwargs.setdefault('verbose_name', kwargs.pop('label'))
        self.other_value = kwargs.pop('other_value', None)
        self.other_label = kwargs.pop('other_label', None)
        self.inner_choices = kwargs.pop('choices', None)
        super().__init__(max_length=max_length, blank=blank, *args, **kwargs)

    def formfield(self, **kwargs):
        return OtherFormField(other_value=self.other_value, other_label=self.other_label, choices=self.inner_choices,
                              label=self.verbose_name, required=not self.blank,
                              **kwargs)


class ListField(models.CharField):
    def __init__(
            self,
            *,
            max_length=10000,
            blank=False,
            **kwargs):

        kwargs.setdefault('help_text', '')
        kwargs.setdefault('null', True)

        super().__init__(
            max_length=max_length,
            blank=blank,
            **kwargs)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if isinstance(value, list):
            return value

        if value is None:
            return value

        return json.loads(value)

    def get_prep_value(self, value):
        return json.dumps(value)


class MultipleChoiceModelField(InnerChoiceMixin, ListField):
    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.inner_choices is None:
            errors.extend([
                checks.Error(
                    "ListField should be provided with choices",
                    obj=self,
                    id='fields.E919',
                )
            ])
        if not isinstance(self.inner_choices, (list, tuple)):
            errors.extend([
                checks.Error(
                    "ListField should be provided with choices parameter: either list or tuple",
                    obj=self,
                    id='fields.E920',
                )
            ])

        default_value = self.initial

        if default_value is not None:
            if not isinstance(default_value, (list, tuple)):
                errors.extend([
                    checks.Error(
                        "Initial value should be either list or tuple",
                        obj=self,
                        id='fields.E921',
                    )
                ])
            inner_choices_flat = [i[0] for i in self.inner_choices]
            if not all(i in inner_choices_flat for i in default_value):
                errors.extend([
                    checks.Error(
                        "Initial values should be item(s) in your 'choices' list",
                        obj=self,
                        id='fields.E922',
                    )
                ])
        return errors

    def __init__(
            self,
            *,
            choices=None,
            max_choices=None,
            min_choices=None,
            label=None,
            max_length=10000,
            doc='',
            initial=None,
            blank=False,
            **kwargs):

        kwargs.update(dict(
            choices=choices,
            label=label,
            initial=initial,
            max_choices=max_choices,
            min_choices=min_choices,
        ))

        self.inner_choices = kwargs.pop('choices', None)
        if self.inner_choices is not None:
            self.max_choices = kwargs.pop('max_choices', len(self.inner_choices))
            self.min_choices = kwargs.pop('min_choices', 0)
        else:
            kwargs.pop('max_choices')
            kwargs.pop('min_choices')
        self.initial = initial
        kwargs.setdefault('help_text', '')
        kwargs.setdefault('null', True)
        label = kwargs['label']
        kwargs.setdefault('verbose_name', label)
        kwargs.setdefault('default', kwargs.pop('initial', None))
        self.initial = kwargs['default']
        if isinstance(self.inner_choices, (list, tuple)):
            self.inner_choices = list(expand_choice_tuples(self.inner_choices))
        kwargs.setdefault('verbose_name', kwargs.pop('label'))
        super().__init__(
            choices=None,
            max_length=max_length,
            blank=None,
            **kwargs)

    def formfield(self, **kwargs):
        if self.inner_choices is not None:
            return MultiSelectFormField(choices=self.inner_choices,
                                        label=self.verbose_name, required=not self.blank,
                                        widget=NoCheckboxCheckbox(choices=self.inner_choices),
                                        max_choices=self.max_choices,
                                        min_choices=self.min_choices,
                                        **kwargs)



import os
import re
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def read_version():
    with open('otree_tools/__init__.py', 'r') as f:
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]",
            f.read(), re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='otree_tools',
    version=read_version(),
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='set of tools facilitating development, exporting and payment management in oTree',
    long_description=README,
    url='https://github.com/chapkovski/otree_tools',
    author='Philipp Chapkovski',
    author_email='chapkovski@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'otree>=2.1.9',
        # 'boto3==1.7.64',
        # 'botocore==1.10.64',
        'django-datetime-widget>=0.9.3',
        'djangorestframework>=3.4.6',
    ]
)



from channels.routing import route_class
from .consumers import TimeTracker, FocusTracker


channel_routing = [
    route_class(TimeTracker, path=TimeTracker.url_pattern, ),
    route_class(FocusTracker, path=FocusTracker.url_pattern, ),
]
